;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)


;; typeins have a 3 pix "border"
;; default-buttons need 4 pix's

(defmethod item-size ((item ccl:simple-view))
  (ccl:view-size item))
(defmethod item-size ((item ccl:editable-text-dialog-item))
  (xy+ (ccl:view-size item) (xy 6 6)))

(defmethod item-position ((item ccl:simple-view))
  (ccl:view-position item))
(defmethod item-position ((item ccl:editable-text-dialog-item))
  (xy- (ccl:view-position item) (xy 3 3)))

(defmethod size-item ((item ccl:simple-view) h &optional v)
  (ccl:set-view-size item (if v (xy h v) h)))
(defmethod size-item ((item ccl:editable-text-dialog-item) h &optional v)
  (ccl:set-view-size item (xy- (if v (xy h v) h) (xy 6 6))))

(defmethod position-item ((item ccl:simple-view)  h &optional v)
  (ccl:set-view-position item (if v (xy h v) h)))
(defmethod position-item ((item ccl:editable-text-dialog-item)
                          h &optional v)
  (ccl:set-view-position item (xy+ (if v (xy h v) h) (xy 3 3))))




(defvar *capella-windows* nil)

(defparameter *capella-font* (make-font :name "Geneva" :size 10))

(defparameter *listing-font* (make-font :name "Courier" :size 10))

(defparameter *bgc*
  (if ccl:*color-available* ccl:*light-gray-color* ccl:*white-color*))

(defconstant +border+ 5)
(defconstant +scroller+ 15)

;;;
;;; resize-mixin causes its superior views to automatically grow or
;;; contract when it is added or removed. the class is currently unused
;;; but table uses the resize-superiors function.
;;;

#|
(defclass resize-mixin () 
  ((mode :initform ':y :initarg :resize-mode))) ; :x,:y,:xy

(defmethod ccl:install-view-in-window :after ((view resize-mixin) 
                                              window)
  (declare (ignore window ))
  (let ((size (ccl:view-size view))
        (mode (slot-value view 'mode)))
    (resize-superiors (ccl:view-container view)
                      (if (or (eq mode ':x) (eq mode ':xy))
                        (xy-x size) 0)
                      (if (or (eq mode ':y) (eq mode ':xy))
                        (xy-y size) 0))))

(defmethod ccl:remove-view-from-window :around ((view resize-mixin))
  (let ((super (ccl:view-container view))
        (size (ccl:view-size view))
        (mode (slot-value view 'mode)))
    (prog1
      (call-next-method)
      (resize-superiors super
                        (if (or (eq mode ':x) (eq mode ':xy))
                          (- (xy-x size)) 0)
                        (if (or (eq mode ':y) (eq mode ':xy))
                          (- (xy-y size)) 0)))))
|#

(defun resize-superiors (sup dx dy)
  (when sup
    (let ((siz (ccl:view-size sup)))
      (ccl:set-view-size sup (xy (+ (xy-x siz) dx)
                                 (+ (xy-y siz) dy)))
      (resize-superiors (ccl:view-container sup) dx dy))))

;;;
;;; capella-window is the root class for all windows.
;;;

(defgeneric update-window-for-changed-object (window object))

(defclass capella-window (ccl:window)
  ()
  (:default-initargs
    :color-p t 
    :close-box-p t
    :window-show nil
    :window-type ':document-with-zoom
    :view-font *capella-font*))

(defmethod ccl:window-close :after ((window capella-window))
  (setf *capella-windows* (delete window *capella-windows*)))

;(defmethod initialize-instance :after ((window capella-window) &rest ignore)
;  (declare (ignore ignore))
;  (push window *capella-windows*))

(defun map-capella-windows (fn &optional (type 'capella-window))
  (map nil #'(lambda (w) (when (typep w type) (funcall fn w)))
       *capella-windows*))

(defmethod update-window-for-changed-object ((browser capella-window) object)
  (declare (ignore object))
  (values))

;;;
;;; window positioning:
;;;   -  sensitive to the worksheet class
;;;   -  initially positions each new window class at a default position
;;;      (4 46) or offset by (18 18) from the position occupied by the
;;;      last capella-window
;;;   -  remembers window positions
;;;   -  stacks windows of the same kind (offsets them by (18 18))
;;;   -  repositions offset windows to (4 46) if they would be off-screen


(let ((window-positions '())
      (default-position (xy 6 44))      ; ccl's standard
      (offset (xy 18 18)))
  (flet ((capella-window-positions ()
           (or window-positions
               (labels ((init-leaf-subclass-positions (class)
                          (let ((sc (slot-value class
                                                'ccl::direct-subclasses)))
                            (if sc
                              (loop for c in sc
                                    do (init-leaf-subclass-positions c))
                              (push (cons class nil)
                                    window-positions)))))
                 (init-leaf-subclass-positions
                  (find-class 'capella-window))))))
   
   
    (defmethod ccl:set-view-position :after ((window capella-window) h
                                             &optional v)
      (setf h (canonicalize-point-args :point h v))
      (setf (cdr (assoc (class-of window) (capella-window-positions))) h)
      ;(call-next-method)
      )
   
    (defmethod initialize-instance :after ((window capella-window)
                                           &rest ignore)
      (declare (ignore ignore))
      (let* ((c (class-of window))
             (w (find-if #'(lambda (x)
                             (typep x c)) *capella-windows*))
             (pos (or
                   ;; if another win of the same kind exists, offset new
                   (when w (xy+ (ccl:view-position w) offset))
                   ;; else use the last position for this kind of win.
                   ;; CAUTION: if this is a newly defined class, first push
                   ;;          it onto window-positions (which returns nil,
                   ;;          so that the next clause applies)
                   (cdr (or (assoc c (capella-window-positions))
                            (assoc c (push (cons c nil) window-positions))))
                   ;; else offset it from the current (capella-) front win
                   (and *capella-windows*
                        (xy+ (ccl:view-position (first *capella-windows*))
                             offset))
                   ;; else use the default position
                   default-position)))
        (when pos
          (ccl:set-view-position window pos)
          ;(print (ccl:window-on-screen-p window))
          (ccl:window-ensure-on-screen window)))
      (push window *capella-windows*)
      window)
    ))

;;;
;;; object-view mixin provides a slot for views that show objects somehow
;;;

(defclass object-view-mixin ()
  ((object :initarg :object :accessor object-view-object :initform nil)))

(defun find-object-view (object view)
  ;; return subview containing object
  (ccl:map-subviews view
                    #'(lambda (v)
                        (when (eq object (object-view-object v))
                          (return-from find-object-view v)))
                    'object-view-mixin)
  nil)

;;;
;;; browser is the root class for building object browsers.
;;;

(defvar *browser* nil "The current active browser.")

(defclass browser (capella-window object-view-mixin)
  ())

(defun browser? (w) (typep w 'browser))

(defmethod ccl:window-close :after ((object browser))
  (when (eq object *browser*) (setf *browser* nil)))

(defmacro with-browser-updated (v &body body)
  `(progn ,@ body (update-browser (view-browser ,v))))

(defmethod ccl:view-activate-event-handler :before ((view browser))
  (setf *browser* view))

(defgeneric browser (object &optional class &rest args))
(defgeneric update-browser (browser))
(defgeneric default-browser-class (object))

(defmethod default-browser-class ((object t))
  nil)
(defmethod default-browser-class ((object standard-object))
  'info-browser)
(defmethod default-browser-class ((object algorithm))
  'algorithm-browser)
(defmethod default-browser-class ((object container)) 
   'listing-browser)
(defmethod default-browser-class ((object generator))
  'listing-browser)

(defun view-browser (view) (ccl:view-window view))

(defun find-browser (object &optional type)
  (find-if #'(lambda (b)
               (and (browser? b)
                    (eq object (object-view-object b))
                    (or (not type) (typep b type))))
           *capella-windows*))


(defmethod browser ((object t) &optional type &rest args)
  (declare (ignore type args))
  (warn-user "~%Browser not implemented for ~S ." object))

(defmethod browser ((object standard-object) &optional class
                    &rest args)
  (ccl:window-select
   (ccl:with-cursor ccl:*watch-cursor*
     (let ((browser 
            (or (find-browser object class)
                (apply #'make-browser 
                       (or class (default-browser-class object))
                       object args))))
       browser))))

(defmethod update-browser ((browser browser)) 
  (values))

(defgeneric set-brower-title (browser &optional object))


(defmacro with-focus-browser ((var) &body forms)
  (let ((v (gensym)))
    `(let ((,v *browser*))
       (when ,v (let ((,var ,v)) ,@ forms)))))

(defun open-browser (&optional type)
  (flet ((doone (o) (browser o type)))
    (with-focus-browser (browser)
      (if (listing-browser? browser)
        (let ((l (browser-listing browser)))
          (if (any-selection? l)
            (map-selected-objects #'(lambda (o) (doone o)) l)
            (doone (object-view-object browser))))
        (doone (object-view-object browser))))))

;;;
;;; Tool-views and tool structure definition.

(defstruct tool
  (name 'no-name :type symbol)
  (action #'(lambda ()) :type function) 
  (icon-id nil :type fixnum))

;(defun find-tool (name list)
;  (find-if #'(lambda (tool) (eq (tool-name tool) name))  list))

;;;
;;; construct a tool view using tool structs. (called from listed.lisp)
;;;

(defun make-tool-view (tools &key view-class view-size view-position
                                rows cols tool-size spacing (offset 0))
  (unless (consp tools) (setf tools (list tools)))
  (let ((nr-of-tools (length tools)))
    (if rows
      (if cols
        ;; adjust too small dimensions, but allow for larger ones
        (when (< (* rows cols) nr-of-tools)
          (multiple-value-bind (timc remc) (ceiling nr-of-tools rows)
            (multiple-value-bind (timr remr) (ceiling nr-of-tools cols)
              ;; remainders should be positiv
              (setf remc (- remc) remr (- remr))
              (if (= remc remr)
                (if (< (- timr rows) (- timc cols)) 
                  (setf rows timr)
                  (setf cols timc))
                (if (< remc remr) 
                  (setf cols timc)
                  (setf rows timr))))))
        (setf cols (ceiling (/ nr-of-tools rows))))
      (if cols
        (setf rows (ceiling (/ nr-of-tools cols)))
        (setf rows 1 cols nr-of-tools))))       ; horizontal by default
  (let ((bgpane (make-instance view-class
                  :view-size 
                  (or view-size
                      ;; two pixel frame!
                      (xy (+ 2 spacing (* (+ tool-size spacing) cols))
                          (+ 2 spacing (* (+ tool-size spacing) rows))))
                  ;; view is framed, so default origin in a window is (-1 -1)
                  :view-position (or view-position (xy 0 0))))
        (size-initarg (xy tool-size tool-size))
        (tool-offset (+ tool-size spacing)))
    (block make-tools
      (loop for r from 1 to rows 
            with tool
            do
            (loop for c from 1 to cols
                  do
                  (when (not tools) (return-from make-tools))
                  (setf tool (pop tools)
                        tool (make-instance 'icon-button
                               :view-nick-name (tool-name tool)
                               :view-size size-initarg
                               :view-position 
                               ;; one pixel left/top frame!
                               (xy (+ 1 spacing (* tool-offset (1- c))
                                      (xy-x offset))
                                   (+ 1 spacing (* tool-offset (1- r))
                                      (xy-y offset)))
                               :view-container bgpane
                               :action (tool-action tool)
                               :icon-id (tool-icon-id tool))))))
    bgpane))


;;;
;;; background color view
;;;

(defclass bgc-mixin ()
  ())

(defmethod ccl:view-draw-contents :around ((view bgc-mixin)) 
  (ccl:with-focused-view view
    (ccl:rlet ((r :rect :topleft (xy 0 0) :bottomright 
                  (ccl:view-size view)))
      (ccl:with-fore-color *bgc*
        (#_fillrect r ccl:*black-pattern*)))
    (call-next-method)))

;;;
;;;
;;;

(defclass 3d-framed-view-mixin ()
  ((borders :initform 4)))

(defmethod ccl:view-draw-contents :before ((view 3d-framed-view-mixin))
  (let* ((size (ccl:view-size view))
         (w (xy-x size))
         (h (xy-y size)))
    (ccl:with-focused-view view
      (frame-3d view (slot-value view 'borders) (1- w) (1- h)))))

(defmethod frame-3d ((me 3d-framed-view-mixin)
                     border width height &optional erase?)
  (let ((left border)
        (top border)
        (right (- width border))
        (bottom (- height border)))
    (%rect-frame-3d left top right bottom :lower erase?)
    (%rect-frame-3d (1+ left) (1+ top) (1- right) (1- bottom) :raise erase?)))


(defmethod ccl:set-view-size :before ((view 3d-framed-view-mixin)
                                      x &optional y)
  (declare (ignore x y))
  (ccl:with-focused-view view
    (let ((size (ccl:view-size view)))
      (frame-3d view (slot-value view 'borders)
                (1- (xy-x size)) (1- (xy-y size)) t))))

(defmethod ccl:set-view-size :after ((view 3d-framed-view-mixin)
                                      x &optional y)
  (declare (ignore x y))
  (ccl:with-focused-view view
    (let ((size (ccl:view-size view)))
      (frame-3d view (slot-value view 'borders)
                (1- (xy-x size)) (1- (xy-y size))))))

(defun %rect-frame-3d (left top right bottom direction erase?)
  (incf bottom 1)                       ; really dunno why...
  (let (c1 c2)
    (if erase?
      (setf c1 *bgc* c2 *bgc*)
      (case direction
        (:raise (setf c1 ccl:*white-color* c2 ccl:*dark-gray-color*))
        (:lower (setf c1 ccl:*dark-gray-color* c2 ccl:*white-color*))))
    (ccl:with-fore-color c1
      (#_moveto left bottom)
      (#_lineto left top)
      (#_lineto right top))
    (ccl:with-fore-color c2
      (#_move 0 1)
      (#_lineto right bottom)
      (#_lineto (1+ left) bottom))))

;;;
;;;
;;;

(defclass gray-3d-framed-view-mixin (bgc-mixin 3d-framed-view-mixin )
  ())

(defclass 3d-framed-view (ccl:view 3d-framed-view-mixin )
  ())

(defclass gray-3d-framed-view (ccl:view gray-3d-framed-view-mixin  )
  ())

(defun canonicalize-point-args (prefer &optional h v)
  (ecase prefer
    (:point (if (and h v)
              (xy h v)
              h))
    (:values (when (and h (not v))
               (setf v (xy-y h)
                     h (xy-x h)))
             (values h v))))

;;;
;;; Worksheets are windows that support activities.
;;;

(defparameter *reuse-worksheets* t)
(defparameter *worksheet-position* '(:top 300))
(defparameter *dxy* (xy 5 5))    ;15 xy border offset for items

(defun worksheet (worksheet)
  (when (symbolp worksheet)
    (setf worksheet (or (and *reuse-worksheets*
                             (find worksheet *capella-windows*
                                   :test #'(lambda (c i) (typep i c))))
                        (make-instance worksheet))))
  (initialize-worksheet worksheet)
  (ccl:window-select worksheet)
  worksheet)

(defmacro with-worksheet-updated (v &body forms)
  `(progn ,@ forms (update-worksheet (view-worksheet ,v))))

(defgeneric worksheet-action (worksheet))
(defgeneric initialize-worksheet (worksheet))
(defgeneric update-worksheet (worksheet))
(defgeneric enable-worksheet (worksheet))
(defgeneric disable-worksheet (worksheet))

(defclass worksheet (capella-window)
  ()
  (:default-initargs
    :window-type :document
    :view-position *worksheet-position* ))

(defclass framed-worksheet (worksheet gray-3d-framed-view-mixin) 
  ())

(defmethod ccl:view-activate-event-handler :after ((view worksheet))
  (initialize-worksheet view))

(defun set-worksheet-title (worksheet prefix browser)
  (ccl:set-window-title 
   worksheet
   (format nil "~A (~A)" prefix
           (object-namestring (object-view-object browser)))))

(defmethod view-worksheet ((view ccl:simple-view))
  (ccl:view-window view))

(defmethod initialize-worksheet ((view worksheet))
  nil)

(defmethod update-worksheet ((view worksheet))
  nil)

(defmethod update-worksheet ((view ccl:dialog-item))
  (update-worksheet (view-worksheet view)))

(defmethod worksheet-action ((view worksheet))
  nil)

(defmethod worksheet-action ((view ccl:dialog-item))
  (worksheet-action (view-worksheet view)))


(defmethod enable-worksheet ((view worksheet))
  nil)

(defmethod disable-worksheet ((view worksheet))
  nil)

;;;
;;; these can be used by worksheets to disable/enable everything
;;;

(defmethod disable-all-dialogs ((window worksheet))
  (ccl:map-subviews window #'control-disable 'ccl:dialog-item))

(defmethod enable-all-dialogs ((window worksheet))
  (ccl:map-subviews window #'control-enable 'ccl:dialog-item))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;;
;;; sequence-view
;;;
;;; 
;;; sequence-view allows for arbitrary vertical scrolling.
;;;   - visible is a cons cell holding the index of the first and
;;;     and last visible entry  
;;;   - view-size is maintained by mcl and equals the view size
;;;     minus scroll bars
;;;   - view-area is the actual used size of the pane, that is,
;;;     view-size + (xy-x scroll-position)
;;;   - cell-size is (xy maximum-estimated-width font-height)
;;; Actual drawing takes place within cells and is clipped to
;;; call-size _and_ frame-size;  highlighting, however, always  
;;; extends horizontally to the vertical scroll bar.
;;; Horizontal scrolling operates as expected (ie, by shifting 
;;; the image); vertical scrolling passes control and responsibility
;;; of drawing to the program.  The vertical scroll bar is set to 
;;; the maximum resolution of #x7fff--<current-position>/#x7fff then 
;;; translates to the a number x/(<nr-of-cells> - <displayable-cells>)


(defconstant +scroll-bar-max+ #x7fff)

(defclass sequence-view (ccl::scroller)
  ((length :initform 0 :accessor sequence-length)
   (sequence :initform nil :initarg :table-sequence 
             :accessor table-sequence)
   (selected :accessor listing-selected :accessor sequence-selected)
   (nselected :initform 0 :accessor sequence-nselected)
   (visible :accessor sequence-visible)
   (cell-size :initarg :cell-size :initform nil
              :accessor sequence-cell-size)
   (view-area :initform 0)
   (last-selected-cell :initform nil :accessor last-selected-cell)))


(defmacro key-arg (key args &optional default)
  `(let ((a (cadr (member ,key ,args))))
     (or a ,default)))

(defmethod initialize-instance :after ((view sequence-view) &rest args)
  (let ((vs (or (key-arg :view-size args)
                      (ccl:view-size view)
                      (ccl:view-default-size view))))
    (setf (slot-value view 'cell-size)
          (xy (- (xy-x vs) 15)
              (font-height *listing-font*)))
    (setf (slot-value view 'view-area) vs)
    ;; call set-table-sequence with slot visible still unbound ...
    (set-table-sequence view (slot-value view 'sequence))
    ;; ... then bind it.
    (setf (slot-value view 'visible) (cons 0 0))
    (setf (ccl:scroll-bar-scroll-size (ccl::v-scroller view))
         
          (let ((invisible (- (sequence-length view)
                              (displayable-lines view #'floor))))  
            (if (> invisible 0)
              (floor (/ +scroll-bar-max+ invisible))
              0)))
    (setf (ccl:scroll-bar-scroll-size (ccl::h-scroller view))
          (xy-x (string-extent "M" (ccl:view-font view))))
    (ccl::update-scroll-bar-limits view)))

(defmethod set-table-sequence ((view sequence-view) sequence)
  (let ((length (length sequence)))
    (setf (slot-value view 'sequence) sequence)
    (setf (slot-value view 'length) length)
    (setf (slot-value view 'selected)
          (make-array length :element-type 'bit
                      :initial-element 0))
    (when (slot-boundp view 'visible)
      (update-sequence-visible view))
    (ccl::update-scroll-bar-limits view)))

(defmacro sequence-first-visible (view)
  `(car (sequence-visible ,view)))

(defmacro sequence-last-visible (view)
  `(cdr (sequence-visible ,view)))

(defun update-sequence-visible (view)
  (let ((lines (ceiling (view-height view)
                        (xy-y (sequence-cell-size view)))))
    (setf (sequence-last-visible view) 
          (min (1- (+ (sequence-first-visible view) lines))
               (1- (sequence-length view))))))
;;
;; helper function;  i use (sequence-cell-size view) just as a matter of 
;; style--if it's not yet set, font-height jumps in

(defun displayable-lines (view round-fn)
  (let ((cell-size (sequence-cell-size view)))
    (funcall round-fn
             (view-height view) (if cell-size 
                                  (xy-y cell-size) 
                                  (font-height *listing-font*)))))

;;
;; scroll-bar limits 
;;   - horizontally limits are 0 and (- (xy-x cell-size) (xy-x view-area)), 
;;     but at least the current horizontal scroll position
;;   - vertical limits are either both 0 or 0 and +scroll-bar-max+

(defmethod ccl::scroll-bar-limits ((view sequence-view))
  ;; make sure cell-size is valid during initialization 
  (let ((cell-size (or (slot-value view 'cell-size) 0)))
    (ccl::normal-scroll-bar-limits view (xy-x cell-size))))


(defmethod ccl::normal-scroll-bar-limits ((view sequence-view) 
                                          max-h &optional max-v)
  (declare (ignore max-v))
  (let* ((size (ccl:view-size view))
         (h-size (xy-x size))
         (x-pos (xy-x (ccl:view-scroll-position view))))
    (setf (slot-value view 'view-area) (xy (+ x-pos h-size) (xy-y size)))
    (values (xy 0 (+ x-pos (max 0 (- max-h h-size))))
            (xy 0 (if (> (slot-value view 'length) 
                         (displayable-lines view #'floor))
                    +scroll-bar-max+
                    0)))))


;;;
;;; gets called by ccl:scroll-bar-changed with arg h bound to a point 

(defmethod ccl::set-view-scroll-position ((view sequence-view) h 
                                          &optional v visibly?)
  (declare (ignore visibly?))
  (multiple-value-setq (h v) (canonicalize-point-args :values h v))
  (let ((pos (ccl::view-scroll-position view))
        (invalidate? nil)
        (new-index (floor (* (- (sequence-length view)
                                (displayable-lines view #'floor))
                             (/ (ccl::scroll-bar-setting 
                                 (slot-value view 'ccl::v-scroller)) 
                                +scroll-bar-max+)))) 
        (old-index (sequence-first-visible view)))
    (unless (= (xy-x pos) h)
      (setf (slot-value view 'ccl:view-scroll-position) (xy h 0))
      (setf (slot-value view 'view-area) 
            (+ (ccl:view-size view) h))
      (call-next-method)
      (setf invalidate? t))
    (unless (= new-index old-index)
      (setf (sequence-first-visible view) new-index)
      (setf invalidate? t)
      (update-sequence-visible view))
    (when invalidate? 
      (ccl::invalidate-view view))
    ;(ccl:view-draw-contents view)
    ;(format t "~&  svsp: (~a ~a) ~a" h v invalidate?)
    (values)))


;; 
;; pass 0 as the vertical scroll bar setting instead of the vertical 
;; scroll position (which is always 0 for scrolling views) when
;; scrolling horizontally

(defun %scroll-sequence-view (view int direction &optional (min 0))
  (ccl::set-view-scroll-position 
     view
     (if (eq direction :vertical)
       (logior (logand #xffff (ccl::view-scroll-position view)) 
               (ash int 16))
       int))
  (when (eql int min)
      (ccl::update-scroll-bar-limits view))
  (ccl::window-update-event-handler (ccl::view-window view)))

(defmethod ccl::scroll-bar-changed ((view sequence-view) scroll-bar)
  (let ((new-value (ccl::scroll-bar-setting scroll-bar)))
    (%scroll-sequence-view view new-value 
                           (slot-value scroll-bar 'ccl::direction)
                           (ccl::scroll-bar-min scroll-bar))))

;;
;; ccl::set-view-scroll-position on ccl::scroller-mixin calls 
;; this after calling the next method--so we need to invalidate
;; it for sequence-views

(defmethod ccl::update-thumbs ((view sequence-view))
  (declare (ignore view)))


;;
;; page size
;; horizontal value is the view area vertical value the displayable
;; lines over the total lines scaled up to +scroll-bar-max+

(defmethod ccl::scroll-bar-page-size ((view sequence-view))
  (xy (xy-x (slot-value view 'view-area))
      (floor (* +scroll-bar-max+ 
                (min 1 
                     (/ (displayable-lines view #'floor)
                        (max 1 (sequence-length view))))))))


;;;
;;; sequence drawing.
;;;

(defun font-base-line (font)
  (multiple-value-bind (a d m l) (ccl:font-info font)
    (declare (ignore d m))
    (+ a l)))


(defmethod ccl:view-draw-contents ((view sequence-view))
  (let ((area (slot-value view 'view-area)))
    (ccl:without-interrupts
      (ccl:with-font-focused-view view
        (ccl:rlet ((rect :rect :left 0 :right (xy-x area)))
          (#_eraserect rect)
          (loop with h = (xy-y (slot-value view 'cell-size))
                with height = (xy-y area)
                with baseline = (font-base-line *listing-font*)
                for i from (sequence-first-visible view) 
                to (sequence-last-visible view)
                for y = 0 then b
                for b = (+ y h)
                do
                (setf (ccl:rref rect rect.top) y)
                (setf (ccl:rref rect rect.bottom) (min b height))
                (draw-table-cell view 0 i rect baseline
                                 (cell-selected-p view 0 i))))))))


(defmethod draw-table-cell ((view sequence-view) 
                            col row rect baseline selected?)
  (ccl:with-clip-rect rect
    (#_EraseRect rect)
    (#_MoveTo (+ (ccl:rref rect rect.left) 3)
     (+ (ccl:rref rect rect.top) baseline))
    (draw-cell-contents view col row)
    (when selected? (highlight-table-cell view nil rect t))))


;;;
;;; col is ignored here; row may be a single row or a list.
;;; the calling pattern is just to comply with ccl's protocol.

(defmethod redraw-cell ((view sequence-view) col &optional row)
  (unless row
    (setf row (xy-y col)))
  (ccl:with-font-focused-view view
    (let* ((area (slot-value view 'view-area))
           (width (xy-x area))
           (height (xy-y area))
           (csize (slot-value view 'cell-size))
           ;(cwidth (xy-x csize))
           (cheight (xy-y csize))
           (baseline (font-base-line *listing-font*))
           (nr-of-elements (slot-value view 'length)))
      (ccl:rlet ((rect :rect :left 0 :right width))
        (flet ((maprow (r)
                 (when (< r nr-of-elements)
                   (let ((top (* (- r (sequence-first-visible view)) cheight)))
                     (setf (ccl:rref rect rect.top) top
                           (ccl:rref rect rect.bottom) (min height 
                                                            (+ top cheight)))
                     (draw-table-cell view 0 r rect baseline
                                      (cell-selected-p view 0 r))))))
          (if (consp row)
            (dolist (r row) (maprow r))
            (maprow row)))))))


(defmethod highlight-table-cell ((view sequence-view) cell
                                 rect selectedp)
  (declare (ignore selectedp cell))
  (#_InvertRect rect))


;;;
;;; mouse selection:
;;; 
;;; - normal click:  toggle selected-p
;;; - command-click: adds or removes single (or multiple, if dragged) 
;;;                  elements to the current selection
;;; - shift-click:   select all elements from the last clicked one to the
;;;                  current element
;;; all modifier keys allow for drag-selection/deselection.  if the drag
;;; speed is too fast, the code catches up with the current mouse position.

(defmethod ccl:view-click-event-handler ((view sequence-view) pos)
  (let ((index (point-to-cell view pos))
        (range (sequence-visible view))
        (length (sequence-length view))
        (redraw '())
        (selected? nil))
    (flet ((track-selection (init fun)
             (declare (integer init) (function fun)
                      (optimize (speed 3) (safety 0)))
             (let* ((lh (xy-y (slot-value view 'cell-size)))
                    (dl (displayable-lines view #'floor))
                    (vh (* dl lh))
                    (len (sequence-length view))
                    (hl (- len dl))
                    (vscr (slot-value view 'ccl::v-scroller))
                    (inc (if (= hl 0)
                           0
                           (floor (/ +scroll-bar-max+ (- len dl)))))
                    (i-1 init))
               (declare (integer lh dl vh len inc i-1))
               (loop for y = (the integer (xy-y (ccl:view-mouse-position
                                                 view)))
                     for i = (the integer (point-to-cell
                                           view 0 (min vh (max y 0))))
                     while (ccl:mouse-down-p)
                     do
                     ;; scroll line by line as long as the mouse is outside
                     (unless (<= 0 y vh)
                       (%scroll-sequence-view view lh :vertical 0)
                       (ccl::set-scroll-bar-setting
                        vscr (funcall (if (< y 0) #'- #'+)
                                      (ccl:scroll-bar-setting vscr) inc)))
                     (unless (or (= i i-1) (>= i len))
                       (funcall fun view 0 i)
                       (setf (last-selected-cell view) i)
                       (when (> (abs (- i i-1)) 1)
                         (let* ((diff (- i i-1))
                                (by (signum diff))
                                (rep (+ (abs diff) by)))
                           (declare (integer diff by rep))
                           (loop for j from (+ i-1 by) by by
                                 repeat rep
                                 do (funcall fun view 0 j)))))
                     (setf i-1 i)
                     finally (unless (>= i len)
                               (setf (last-selected-cell view) i))))))
      (when (< index length)
        (setf selected? (cell-selected-p view 0 index)))
      (cond ((ccl:command-key-p)
             (when (< index length)
               (let ((fun (if selected? #'cell-deselect #'cell-select)))
                 (funcall fun view 0 index)
                 (track-selection index fun))))
            ((ccl:shift-key-p)
             (setf index (min (1- (sequence-length view)) 
                              (max index 0)))
             (let* ((last (or (last-selected-cell view) index))
                    (s (min index last))
                    (e (max index last)))
               (setf redraw
                     (loop for i from (car range) to (cdr range)
                           for p = (cell-selected-p view 0 i)
                           when (if (<= s i e) 
                                  (not p)
                                  p)
                           collect i))
               (loop for i from s to e
                     unless (cell-selected-p view 0 i)
                     do (%cell-select view 0 i))
               (funcall #'redraw-cell view 0 redraw)
               (track-selection index #'cell-select)))
            (t 
             (when t ;(not selected?)
               (when (> (sequence-nselected view) 0)
                 (setf redraw 
                       (loop for i from (car range) to (cdr range)
                             when (cell-selected-p view 0 i)
                             collect i))
                 (%deselect-all-cells view))
               (when (< index (sequence-length view))
                 (%cell-select view 0 index)
                 (pushnew index redraw))
               (funcall #'redraw-cell view 0 redraw))
             (track-selection index #'cell-select))))))

;;
;; point2cell
;; Since we never scroll the view vertically, vertical coords are always 0.
;; Thus, add (sequence-first-visible view) to get the actual cell.

(defmethod point-to-cell ((view sequence-view) xy &optional y)
  (+ (sequence-first-visible view)
     (floor (or y (xy-y xy)) (xy-y (slot-value view 'cell-size)))))


;; unused
;;
;(defmethod table-dimensions ((view sequence-view))
;  (xy 0 (sequence-length view)))

(defmethod cell-selected-p ((view sequence-view) col &optional row)
  (= (the bit
       (aref (the simple-bit-vector (slot-value view 'selected))
             (the integer (or row (xy-y col)))))
     1))

(defmacro %%set-cell-selection (array row value)
  `(setf (aref (the simple-bit-vector ,array)
               (the integer ,row))
         (the bit ,value)))

(defmethod cell-select ((view sequence-view) col &optional row)
  (unless (cell-selected-p view col row) 
    (%cell-select view col row)
    (redraw-cell view col row)))

(defmethod cell-deselect ((view sequence-view) col &optional row)
  (when (cell-selected-p view col row) 
    (%cell-deselect view col row)
    (redraw-cell view col row)))

(defmethod %cell-select ((view sequence-view) col &optional row)
  (%%set-cell-selection (slot-value view 'selected) 
                        (or row (xy-y col))
                        1)
  (incf (slot-value view 'nselected)))

(defmethod %cell-deselect ((view sequence-view) col &optional row)
  (%%set-cell-selection (slot-value view 'selected) 
                        (or row (xy-y col))
                        0)
  (decf (slot-value view 'nselected)))

(defmethod selected-cells ((view sequence-view))
  (let ((v (slot-value view 'selected)) 
        (s (slot-value view 'nselected))
        (c 0))
    (declare (simple-bit-vector v))
    (loop for i below (sequence-length view)
          while (< c s)
          when (= (the bit (aref v i)) 1)
          collect i and do (incf c))))

(defmethod map-selected-cells (fn (view sequence-view))
  (let ((v (sequence-selected view))
        (l (sequence-length view))
        (s (sequence-nselected view))
        (c 0))
    (declare (simple-bit-vector v) (integer l s c))
    (loop for i below l
          while (< c s)
          when (= (the bit (aref v i)) 1)
          do
          (funcall fn i)
          (incf c))
    (values)))

(defmethod map-selected-objects (fn (view sequence-view))
  (let ((v (sequence-selected view))
        (s (table-sequence view))
        (n (sequence-nselected view))
        (c 0)
        (i 0))
    (declare (simple-bit-vector v) (integer n c i) (list s)
             ;(optimize (speed 3 safety 0))
             )
    (loop while (< c n)
          do
          (when (= (the bit (aref v i)) 1)
            (funcall fn (car s))
            (incf c))
          (incf i)
          (setf s (cdr s)))
    (values)))

(defun selected-objects (sequence)
  (let ((l '()))
    (map-selected-objects #'(lambda (o) (push o l)) sequence)
    (nreverse l)))

(defmethod map-sequence-cells (fn (view sequence-view))
  (loop for i below (sequence-length view) do (funcall fn i)))

(defmethod map-sequence-objects (fn (view sequence-view))
  (loop for o in (table-sequence view) do (funcall fn o)))

(defmethod select-all-cells ((view sequence-view))
  (%select-all-cells view)
  (ccl:invalidate-view view))

(defmethod deselect-all-cells ((view sequence-view))
  (%deselect-all-cells view)
  (ccl:invalidate-view view))

(defmethod %select-all-cells (view)
  (let ((v (sequence-selected view)))
    (loop for i below (sequence-length view)
          do 
          (%%set-cell-selection v i 1))
    (setf (slot-value view 'nselected) (sequence-length view))))

(defun %deselect-all-cells (view)
  (let ((v (sequence-selected view))
        (l (sequence-length view))
        (c 0))
    (declare (simple-bit-vector v) (integer l s c))
    (loop for i below l
          do
          (%%set-cell-selection v i 0)
          (incf c))
    (setf (slot-value view 'nselected) 0)))

(defun first-selected-cell (view &optional from-end)
  (and (> (sequence-nselected view) 0)
       (position 1 (sequence-selected view) :from-end from-end)))


;;;
;;; default cell functions.
;;;

(defmethod draw-cell-contents ((view sequence-view) col &optional row)
  (declare (ignore col))
  (ccl:with-pstrs ((s (prin1-to-string (cell-contents view 0 row))))
    (#_DrawString s)))


(defmethod cell-contents ((view sequence-view) col &optional row)
  ;; default cell contents funtion
  (declare (ignore col))
  (elt (table-sequence view) row))

;;;
;;; COMMAND-U outut processing for capella windows
;;;

(defgeneric window-send-output (w &optional s o))
(defgeneric window-send-output? (w &optional s check?))
(defparameter *default-output-start* .5)

(defmethod window-send-output (w &optional (s *output-stream*)
                                 (o *default-output-start*))
  (declare (ignore w s o))
  (ccl::beep))

(defmethod window-send-output ((w capella-window)
                               &optional (s *output-stream*)
                               (o *default-output-start*))
  (if s
    (let ((x (window-send-output? w s nil)))
      (if x
        (let ((m (if (consp x)
                   (every #'(lambda (i) (typep i 'startable-element)) x)
                   (not (event? x))))
              (*post-processing* nil))
          (if m 
            (mix-object x s o)
            (seq-object x s o)))
        (warn-user "Can't output: no objects selected.")))
    (warn-user "Can't output: no output stream set.")))

(defmethod window-send-output? (w &optional s check?)
  (declare (ignore w s check?))
  nil)

(defmethod window-send-output? ((w browser)
                                &optional (s *browser*) (check? t)) 
  ;; return nil, t or selected objects if not check.
  (declare (special %check-selections-limit%))
  (if (and w s)
    (let ((o (object-view-object w)))
      (if (eq o .top-level.)
        (if (any-selection? w) 
          (or check? (selected-objects (browser-listing w))))
        (if (systemp o)
          nil
          (if (listing-browser? w)
            (let ((l (browser-listing w)))
              (if (any-selection? l)
                (and (selection-class? l %check-selections-limit% 
                                       'timed-object)
                     (or check? (selected-objects l)))
                (and (typep o 'timed-object)
                     (or check? o))))
            (and (typep o 'timed-object) (or check? o))))))
    nil))

