;;;
;;;
;;; File:         pmcl-entry.lisp
;;;
;;; Summary:      Shared-Library entry points of the reimplementation
;;;               of the Apple MidiManager interface for the PPC version of 
;;;               Macintosh Common Lisp (3.9 or higher).
;;;
;;; Dependencies: pmcl-midi.cc
;;;               pmcl-midi.hh
;;;               pmcl-midi.lisp
;;;
;;;               ***********************************************************
;;;               Copyright (c) 1996 by Tobias Kunze and Rick Taube.  All
;;;               rights reserved.  Use and copying of this software and
;;;               preparation of derivative works based upon this software
;;;               are permitted and may be copied as long as no fees or
;;;               compensation are charged for use, copying, or accessing
;;;               this software and all copies of this software include this
;;;               copyright notice.  Suggestions, comments and bug reports
;;;               are welcome.  Please address email regarding this midi
;;;               implementation to: t@kunze.stanford.edu or
;;;               hkt@ccrma.stanford.edu.
;;;               ***********************************************************
;;;
;;; Author:       Tobias Kunze
;;; e-Mail:       t@kunze.stanford.edu
;;; Org:          CCRMA, Stanford University
;;;
;;; orig-Date:    20-Jul-96 at 02:31:36
;;; Last-Mod:     26-Jul-96 at 00:13:46 by Tobias Kunze
;;;
;;;
;;; Description:  This code, together with the files pmcl-midi.cc, 
;;;               pmcl-midi.lisp and pmcl-midi.hh interfaces with the Apple 
;;;               Midi Manager.  It is intended to be part of Common Music by 
;;;               Rick Taube but may be used in standalone applications as 
;;;               well.  To use it without Common Music, follow the 
;;;               instructions in the section CLIENT DEFINITIONS below.
;;;
;;; Revision:     1.1
;;;
;;; Changes:
;;;   18-Jul-96   hkt  moved contents of this file from pmcl-midi.lisp to 
;;;                    insure that these get defined before this file is 
;;;                    compiled. 
;;;   24-Jul-95   tk   added unwind-protect to read-hook


;;; -------------------------------------------------------------------------
;;;
;;; CLIENT DEFINITIONS
;;;

(in-package #+cm :common-music #-cm :cl-user)

;;;
;;; END CLIENT DEFINITIONS
;;;
;;; -------------------------------------------------------------------------


;;; -------------------------------------------------------------------------
;;;
;;; C Entry Points. This file MUST be loaded before functions that use
;;; these entry points are COMPILED.
;;;
;;;

;;;
;;;   +pmcl-midi-shlib+                          <default: "pmcl-midi.shlib">
;;;      The name of our shared library.

(defconstant +pmcl-midi-shlib+ "pmcl-midi.shlib")


(eval-when (:load-toplevel :execute)
  (ccl::add-to-shared-library-search-path +pmcl-midi-shlib+))

;;;
;;; Register a port's reference number with the C side after we allocated it.

(ccl:define-entry-point (RegisterInvInPort      ; +pmcl-midi-shlib+
			 ("pmcl-midi.shlib" "RegisterInvInPort__Fs"))
  ((refnum :short))
  :short)

;;;
;;; Rotate the C side refnums left by idx places.  By convention, we always
;;; write to the first refnum.

(ccl:define-entry-point ("RotateInvInPorts"       ; +pmcl-midi-shlib+
			 ("pmcl-midi.shlib" "RotateInvInPorts__Fi"))
  ((idx :short))
  nil)

;;;
;;; This goes down to C and actually deallocates the ports whose refnums have
;;; been moved to Queues.fObsInPorts[]

(ccl:define-entry-point ("RemoveObsPorts"        ; +pmcl-midi-shlib+
                         ("pmcl-midi.shlib" "RemoveObsPorts__Fv"))
  nil
  nil)

;;;
;;; Poll the input queues.

(ccl:define-entry-point ("MidiReadEvent"        ; +pmcl-midi-shlib+
                         ("pmcl-midi.shlib" "MidiReadEvent__Fv"))
  nil
  :short)


;;;
;;; Set the ReadHook field of the (newly created) invisible Input port with
;;; reference number refnum to InvReadHook().

(ccl:define-entry-point ("SetReadHook"        ; +pmcl-midi-shlib+
                         ("pmcl-midi.shlib" "SetReadHook__Fs"))
  ((refnum :short))
  nil)

;;;
;;; Flush Output.  This sets the invisible read hook to flushing mode and
;;; automatically back after a while (see *midi-flush-timeout* above).

(ccl:define-entry-point ("FlushQueues"        ; +pmcl-midi-shlib+
                         ("pmcl-midi.shlib" "FlushQueues__Fv"))
  nil
  nil)

;;;
;;; The LISP side midi read hook.  

(ccl:defpascal %midi-read-hook% (:long msg :long tim :void)
  (declare (special *midi-read-hook*)
           (optimize (speed 3) (safety 0)))
  (when *midi-read-hook*
    (block :read-hook
      (let ((done? nil))
        (unwind-protect 
          (progn 
            (funcall (the function *midi-read-hook*)
                     (the fixnum msg) (the fixnum tim))
            (setf done? t))
          (unless done?
            (setf *midi-read-hook* nil)
            (warn "Abort in user-supplied *midi-read-hook*.  Bailing out.")
            (return-from :read-hook nil)))))))

;;;
;;; Holds the transition vector of the %midi-read-hook% function

(defparameter %lisp-midi-read-hook% nil)

;;;
;;; C Side Initialization.

(ccl:define-entry-point ("MidiCSetUp"        ; +pmcl-midi-shlib+
                         ("pmcl-midi.shlib" "MidiCSetUp__FssssssPFll_vlll"))
  ((time :short) (inA :short) (inB :short) (outA :short) (outB :short) 
   (invout :short)
   (readhook :ptr) (wakeupdur :long) (buftimeout :long) (flushdur :long))
  :ptr)

;;;
;;; Get the adress of bOutputInProgress

(ccl:define-entry-point ("GetbOutputInProgress"       ; +pmcl-midi-shlib+
			 ("pmcl-midi.shlib" "GetbOutputInProgress__Fv"))
  nil
  :ptr)



;;;
;;; -*- EOF -*-
